@@ -26,7 +26,7 @@
 
 
 #include "cpio_platform.h"
-__FBSDID("$FreeBSD$");
+__FBSDID("$FreeBSD: src/usr.bin/cpio/cmdline.c,v 1.3 2008/06/21 02:20:20 kientzle Exp $");
 
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
@@ -66,7 +66,7 @@ struct option {
  */
 
 
-static const char *cpio_opts = "aBcdf:H:hijLlmopR:rtuvW:yz";
+static const char *cpio_opts = "0AaBC:F:O:cdE:f:H:hijLlmopR:rtuvW:yZz";
 
 /*
  * On systems that lack getopt_long, long options can be specified
@@ -82,10 +82,21 @@ static const char *cpio_opts = "aBcdf:H:hijLlmopR:rtuvW:yz";
  * sorted, as the -W logic below relies on it.
  */
 static const struct option cpio_longopts[] = {
+	{ "create",		no_argument,	   NULL, 'o' },
+	{ "extract",		no_argument,       NULL, 'i' },
+	{ "file",		required_argument, NULL, 'F' },
 	{ "format",             required_argument, NULL, 'H' },
 	{ "help",		no_argument,	   NULL, 'h' },
+	{ "insecure",		no_argument,	   NULL, OPTION_INSECURE },
+	{ "link",		no_argument,	   NULL, 'l' },
+	{ "list",		no_argument,	   NULL, 't' },
+	{ "make-directories",	no_argument,	   NULL, 'd' },
+	{ "null",		no_argument,	   NULL, '0' },
 	{ "owner",		required_argument, NULL, 'R' },
+	{ "pass-through",	no_argument,	   NULL, 'p' },
+	{ "preserve-modification-time", no_argument, NULL, 'm' },
 	{ "quiet",		no_argument,	   NULL, OPTION_QUIET },
+	{ "unconditional",	no_argument,	   NULL, 'u' },
 	{ "verbose",            no_argument,       NULL, 'v' },
 	{ "version",            no_argument,       NULL, OPTION_VERSION },
 	{ NULL, 0, NULL, 0 }
