@@ -26,7 +26,7 @@
 
 
 #include "cpio_platform.h"
-__FBSDID("$FreeBSD$");
+__FBSDID("$FreeBSD: src/usr.bin/cpio/cpio.c,v 1.4 2008/06/24 15:18:40 kientzle Exp $");
 
 #include <sys/types.h>
 #include <archive.h>
@@ -60,23 +60,22 @@ __FBSDID("$FreeBSD$");
 
 static int	copy_data(struct archive *, struct archive *);
 static const char *cpio_rename(const char *name);
-static int	file_to_archive(struct cpio *, const char *, const char *);
+static int	entry_to_archive(struct cpio *, struct archive_entry *);
+static int	file_to_archive(struct cpio *, const char *);
 static void	long_help(void);
 static void	mode_in(struct cpio *);
 static void	mode_list(struct cpio *);
 static void	mode_out(struct cpio *);
 static void	mode_pass(struct cpio *, const char *);
-static int	out_file(struct cpio *, const char *pathname);
-static int	process_lines(struct cpio *cpio, const char *pathname,
-		    int (*process)(struct cpio *, const char *));
 static void	restore_time(struct cpio *, struct archive_entry *,
 		    const char *, int fd);
 static void	usage(void);
-static void	version(FILE *);
+static void	version(void);
 
 int
 main(int argc, char *argv[])
 {
+	static char buff[16384];
 	struct cpio _cpio; /* Allocated on stack. */
 	struct cpio *cpio;
 	int uid, gid;
@@ -84,6 +83,8 @@ main(int argc, char *argv[])
 
 	cpio = &_cpio;
 	memset(cpio, 0, sizeof(*cpio));
+	cpio->buff = buff;
+	cpio->buff_size = sizeof(buff);
 
 	/* Need cpio_progname before calling cpio_warnc. */
 	if (*argv == NULL)
@@ -100,6 +101,7 @@ main(int argc, char *argv[])
 	cpio->gid_override = -1;
 	cpio->argv = argv;
 	cpio->argc = argc;
+	cpio->line_separator = '\n';
 	cpio->mode = '\0';
 	cpio->verbose = 0;
 	cpio->compress = '\0';
@@ -107,37 +109,63 @@ main(int argc, char *argv[])
 	cpio->format = "odc"; /* Default format */
 	cpio->extract_flags = ARCHIVE_EXTRACT_NO_AUTODIR;
 	cpio->extract_flags |= ARCHIVE_EXTRACT_NO_OVERWRITE_NEWER;
+	cpio->extract_flags |= ARCHIVE_EXTRACT_SECURE_SYMLINKS;
+	cpio->extract_flags |= ARCHIVE_EXTRACT_SECURE_NODOTDOT;
 	/* TODO: If run by root, set owner as well. */
 	cpio->bytes_per_block = 512;
 	cpio->filename = NULL;
 
 	while ((opt = cpio_getopt(cpio)) != -1) {
 		switch (opt) {
+		case '0': /* GNU convention: --null, -0 */
+			cpio->line_separator = '\0';
+			break;
+		case 'A': /* NetBSD/OpenBSD */
+			cpio->option_append = 1;
+			break;
 		case 'a': /* POSIX 1997 */
 			cpio->option_atime_restore = 1;
 			break;
 		case 'B': /* POSIX 1997 */
 			cpio->bytes_per_block = 5120;
 			break;
+		case 'C': /* NetBSD/OpenBSD */
+			cpio->bytes_per_block = atoi(optarg);
+			if (cpio->bytes_per_block <= 0)
+				cpio_errc(1, 0, "Invalid blocksize %s", optarg);
+			break;
 		case 'c': /* POSIX 1997 */
 			cpio->format = "odc";
 			break;
 		case 'd': /* POSIX 1997 */
 			cpio->extract_flags &= ~ARCHIVE_EXTRACT_NO_AUTODIR;
 			break;
+		case 'E': /* NetBSD/OpenBSD */
+			include_from_file(cpio, optarg);
+			break;
+		case 'F': /* NetBSD/OpenBSD/GNU cpio */
+			cpio->filename = optarg;
+			break;
 		case 'f': /* POSIX 1997 */
 			exclude(cpio, optarg);
 			break;
-		case 'H': /* GNU cpio, also --format */
+		case 'H': /* GNU cpio (also --format) */
 			cpio->format = optarg;
 			break;
 		case 'h':
 			long_help();
 			break;
+		case 'I': /* NetBSD/OpenBSD */
+			cpio->filename = optarg;
+			break;
 		case 'i': /* POSIX 1997 */
 			cpio->mode = opt;
 			break;
-		case 'L': /* GNU cpio, BSD convention */
+		case OPTION_INSECURE:
+			cpio->extract_flags &= ~ARCHIVE_EXTRACT_SECURE_SYMLINKS;
+			cpio->extract_flags &= ~ARCHIVE_EXTRACT_SECURE_NODOTDOT;
+			break;
+		case 'L': /* GNU cpio */
 			cpio->option_follow_links = 1;
 			break;
 		case 'l': /* POSIX 1997 */
@@ -146,11 +174,15 @@ main(int argc, char *argv[])
 		case 'm': /* POSIX 1997 */
 			cpio->extract_flags |= ARCHIVE_EXTRACT_TIME;
 			break;
+		case 'O': /* GNU cpio */
+			cpio->filename = optarg;
+			break;
 		case 'o': /* POSIX 1997 */
 			cpio->mode = opt;
 			break;
 		case 'p': /* POSIX 1997 */
 			cpio->mode = opt;
+			cpio->extract_flags &= ~ARCHIVE_EXTRACT_SECURE_NODOTDOT;
 			break;
 		case OPTION_QUIET: /* GNU cpio */
 			cpio->quiet = 1;
@@ -177,7 +209,7 @@ main(int argc, char *argv[])
 			cpio->verbose++;
 			break;
 		case OPTION_VERSION: /* GNU convention */
-			version(stdout);
+			version();
 			break;
 #if 0
 	        /*
@@ -190,6 +222,9 @@ main(int argc, char *argv[])
 		case 'y': /* tar convention */
 			cpio->compress = opt;
 			break;
+		case 'Z': /* tar convention */
+			cpio->compress = opt;
+			break;
 		case 'z': /* tar convention */
 			cpio->compress = opt;
 			break;
@@ -296,24 +331,29 @@ long_help(void)
 		} else
 			putchar(*p);
 	}
-	version(stdout);
+	version();
 }
 
 static void
-version(FILE *out)
+version(void)
 {
-	fprintf(out,"bsdcpio %s -- %s\n",
+	fprintf(stdout,"bsdcpio %s -- %s\n",
 	    BSDCPIO_VERSION_STRING,
 	    archive_version());
-	exit(1);
+	exit(0);
 }
 
 static void
 mode_out(struct cpio *cpio)
 {
 	unsigned long blocks;
+	struct archive_entry *entry, *spare;
+	struct line_reader *lr;
+	const char *p;
 	int r;
 
+	if (cpio->option_append)
+		cpio_errc(1, 0, "Append mode not yet supported.");
 	cpio->archive = archive_write_new();
 	if (cpio->archive == NULL)
 		cpio_errc(1, 0, "Failed to allocate archive object");
@@ -324,6 +364,9 @@ mode_out(struct cpio *cpio)
 	case 'z':
 		archive_write_set_compression_gzip(cpio->archive);
 		break;
+	case 'Z':
+		archive_write_set_compression_compress(cpio->archive);
+		break;
 	default:
 		archive_write_set_compression_none(cpio->archive);
 		break;
@@ -333,11 +376,29 @@ mode_out(struct cpio *cpio)
 		cpio_errc(1, 0, archive_error_string(cpio->archive));
 	archive_write_set_bytes_per_block(cpio->archive, cpio->bytes_per_block);
 	cpio->linkresolver = archive_entry_linkresolver_new();
+	archive_entry_linkresolver_set_strategy(cpio->linkresolver,
+	    archive_format(cpio->archive));
 
 	r = archive_write_open_file(cpio->archive, cpio->filename);
 	if (r != ARCHIVE_OK)
 		cpio_errc(1, 0, archive_error_string(cpio->archive));
-	process_lines(cpio, "-", out_file);
+	lr = process_lines_init("-", cpio->line_separator);
+	while ((p = process_lines_next(lr)) != NULL)
+		file_to_archive(cpio, p);
+	process_lines_free(lr);
+
+	/*
+	 * The hardlink detection may have queued up a couple of entries
+	 * that can now be flushed.
+	 */
+	entry = NULL;
+	archive_entry_linkify(cpio->linkresolver, &entry, &spare);
+	while (entry != NULL) {
+		entry_to_archive(cpio, entry);
+		archive_entry_free(entry);
+		entry = NULL;
+		archive_entry_linkify(cpio->linkresolver, &entry, &spare);
+	}
 
 	r = archive_write_close(cpio->archive);
 	if (r != ARCHIVE_OK)
@@ -352,74 +413,39 @@ mode_out(struct cpio *cpio)
 	archive_write_finish(cpio->archive);
 }
 
-static int
-out_file(struct cpio *cpio, const char *path)
-{
-	const char *destpath;
-
-	if (cpio->option_rename)
-		destpath = cpio_rename(path);
-	else
-		destpath = path;
-	if (destpath == NULL)
-		return (0);
-	return (file_to_archive(cpio, path, destpath));
-}
-
 /*
  * This is used by both out mode (to copy objects from disk into
  * an archive) and pass mode (to copy objects from disk to
  * an archive_write_disk "archive").
  */
 static int
-file_to_archive(struct cpio *cpio, const char *srcpath, const char *destpath)
+file_to_archive(struct cpio *cpio, const char *srcpath)
 {
-	static char buff[16384];
 	struct stat st;
-	struct archive_entry *entry;
-	const char *linkname;
-	int fd = -1;
-	ssize_t len;
+	const char *destpath;
+	struct archive_entry *entry, *spare;
+	size_t len;
+	const char *p;
+	int lnklen;
 	int r;
 
-	if (cpio->verbose)
-		fprintf(stderr,"%s", destpath);
-
+	/*
+	 * Create an archive_entry describing the source file.
+	 */
 	entry = archive_entry_new();
 	if (entry == NULL)
 		cpio_errc(1, 0, "Couldn't allocate entry");
-	archive_entry_set_pathname(entry, destpath);
-
-	/* TODO: pathname editing. */
+	archive_entry_copy_sourcepath(entry, srcpath);
 
+	/* Get stat information. */
 	if (cpio->option_follow_links)
 		r = stat(srcpath, &st);
 	else
 		r = lstat(srcpath, &st);
 	if (r != 0) {
 		cpio_warnc(errno, "Couldn't stat \"%s\"", srcpath);
-		goto cleanup;
-	}
-
-	/* If we're trying to preserve hardlinks, match them here. */
-	if (cpio->linkresolver != NULL
-	    && st.st_nlink > 1
-	    && !S_ISDIR(st.st_mode)) {
-		linkname = archive_entry_linkresolve(cpio->linkresolver, entry);
-		archive_entry_set_hardlink(entry, linkname);
-	}
-
-	if (S_ISLNK(st.st_mode)) {
-		int lnklen;
-
-		lnklen = readlink(srcpath, buff, sizeof(buff));
-		if (lnklen < 0) {
-			cpio_warnc(errno,
-			    "%s: Couldn't read symbolic link", srcpath);
-			goto cleanup;
-		}
-		buff[lnklen] = 0;
-		archive_entry_set_symlink(entry, buff);
+		archive_entry_free(entry);
+		return (0);
 	}
 
 	if (cpio->uid_override >= 0)
@@ -428,17 +454,123 @@ file_to_archive(struct cpio *cpio, const char *srcpath, const char *destpath)
 		st.st_gid = cpio->uid_override;
 	archive_entry_copy_stat(entry, &st);
 
-	/* Obviously, this only gets invoked in pass mode. */
-	if (cpio->option_link) {
-		/* Note: link(2) doesn't create parent directories. */
-		archive_entry_set_hardlink(entry, srcpath);
-		r = archive_write_header(cpio->archive, entry);
-		if (r == ARCHIVE_OK)
+	/* If its a symlink, pull the target. */
+	if (S_ISLNK(st.st_mode)) {
+		lnklen = readlink(srcpath, cpio->buff, cpio->buff_size);
+		if (lnklen < 0) {
+			cpio_warnc(errno,
+			    "%s: Couldn't read symbolic link", srcpath);
+			archive_entry_free(entry);
 			return (0);
-		cpio_warnc(archive_errno(cpio->archive),
-		    archive_error_string(cpio->archive));
+		}
+		cpio->buff[lnklen] = 0;
+		archive_entry_set_symlink(entry, cpio->buff);
+	}
+
+	/*
+	 * Generate a destination path for this entry.
+	 * "destination path" is the name to which it will be copied in
+	 * pass mode or the name that will go into the archive in
+	 * output mode.
+	 */
+	destpath = srcpath;
+	if (cpio->destdir) {
+		len = strlen(cpio->destdir) + strlen(srcpath) + 8;
+		if (len >= cpio->pass_destpath_alloc) {
+			while (len >= cpio->pass_destpath_alloc) {
+				cpio->pass_destpath_alloc += 512;
+				cpio->pass_destpath_alloc *= 2;
+			}
+			free(cpio->pass_destpath);
+			cpio->pass_destpath = malloc(cpio->pass_destpath_alloc);
+			if (cpio->pass_destpath == NULL)
+				cpio_errc(1, ENOMEM,
+				    "Can't allocate path buffer");
+		}
+		strcpy(cpio->pass_destpath, cpio->destdir);
+		p = srcpath;
+		while (p[0] == '/')
+			++p;
+		strcat(cpio->pass_destpath, p);
+		destpath = cpio->pass_destpath;
+	}
+	if (cpio->option_rename)
+		destpath = cpio_rename(destpath);
+	if (destpath == NULL)
+		return (0);
+	archive_entry_copy_pathname(entry, destpath);
+
+	/*
+	 * If we're trying to preserve hardlinks, match them here.
+	 */
+	spare = NULL;
+	if (cpio->linkresolver != NULL
+	    && !S_ISDIR(st.st_mode)) {
+		archive_entry_linkify(cpio->linkresolver, &entry, &spare);
+	}
+
+	if (entry != NULL) {
+		r = entry_to_archive(cpio, entry);
+		archive_entry_free(entry);
+	}
+	if (spare != NULL) {
+		if (r == 0)
+			r = entry_to_archive(cpio, spare);
+		archive_entry_free(spare);
+	}
+	return (r);
+}
+
+static int
+entry_to_archive(struct cpio *cpio, struct archive_entry *entry)
+{
+	const char *destpath = archive_entry_pathname(entry);
+	const char *srcpath = archive_entry_sourcepath(entry);
+	int fd = -1;
+	ssize_t bytes_read;
+	int r;
+
+	/* Print out the destination name to the user. */
+	if (cpio->verbose)
+		fprintf(stderr,"%s", destpath);
+
+	/*
+	 * Option_link only makes sense in pass mode and for
+	 * regular files.  Also note: if a link operation fails
+	 * because of cross-device restrictions, we'll fall back
+	 * to copy mode for that entry.
+	 *
+	 * TODO: Test other cpio implementations to see if they
+	 * hard-link anything other than regular files here.
+	 */
+	if (cpio->option_link
+	    && archive_entry_filetype(entry) == AE_IFREG)
+	{
+		struct archive_entry *t;
+		/* Save the original entry in case we need it later. */
+		t = archive_entry_clone(entry);
+		if (t == NULL)
+			cpio_errc(1, ENOMEM, "Can't create link");
+		/* Note: link(2) doesn't create parent directories,
+		 * so we use archive_write_header() instead as a
+		 * convenience. */
+		archive_entry_set_hardlink(t, srcpath);
+		/* This is a straight link that carries no data. */
+		archive_entry_set_size(t, 0);
+		r = archive_write_header(cpio->archive, t);
+		archive_entry_free(t);
+		if (r != ARCHIVE_OK)
+			cpio_warnc(archive_errno(cpio->archive),
+			    archive_error_string(cpio->archive));
 		if (r == ARCHIVE_FATAL)
 			exit(1);
+#ifdef EXDEV
+		if (r != ARCHIVE_OK && archive_errno(cpio->archive) == EXDEV) {
+			/* Cross-device link:  Just fall through and use
+			 * the original entry to copy the file over. */
+			cpio_warnc(0, "Copying file instead");
+		} else
+#endif
 		return (0);
 	}
 
@@ -446,17 +578,18 @@ file_to_archive(struct cpio *cpio, const char *srcpath, const char *destpath)
 	 * Make sure we can open the file (if necessary) before
 	 * trying to write the header.
 	 */
-	if (S_ISREG(st.st_mode) && st.st_size > 0) {
-		fd = open(srcpath, O_RDONLY);
-		if (fd < 0) {
-			cpio_warnc(errno,
-			    "%s: could not open file", srcpath);
-			goto cleanup;
+	if (archive_entry_filetype(entry) == AE_IFREG) {
+		if (archive_entry_size(entry) > 0) {
+			fd = open(srcpath, O_RDONLY);
+			if (fd < 0) {
+				cpio_warnc(errno,
+				    "%s: could not open file", srcpath);
+				goto cleanup;
+			}
 		}
-	}
-
-	if (!S_ISREG(st.st_mode))
+	} else {
 		archive_entry_set_size(entry, 0);
+	}
 
 	r = archive_write_header(cpio->archive, entry);
 
@@ -469,19 +602,19 @@ file_to_archive(struct cpio *cpio, const char *srcpath, const char *destpath)
 	if (r == ARCHIVE_FATAL)
 		exit(1);
 
-	if (r >= ARCHIVE_WARN && fd >= 0 && archive_entry_size(entry) > 0) {
-		fd = open(srcpath, O_RDONLY);
-		len = read(fd, buff, sizeof(buff));
-		while (len > 0) {
-			r = archive_write_data(cpio->archive, buff, len);
+	if (r >= ARCHIVE_WARN && fd >= 0) {
+		bytes_read = read(fd, cpio->buff, cpio->buff_size);
+		while (bytes_read > 0) {
+			r = archive_write_data(cpio->archive,
+			    cpio->buff, bytes_read);
 			if (r < 0)
 				cpio_errc(1, archive_errno(cpio->archive),
 				    archive_error_string(cpio->archive));
-			if (r < len) {
+			if (r < bytes_read) {
 				cpio_warnc(0,
 				    "Truncated write; file may have grown while being archived.");
 			}
-			len = read(fd, buff, sizeof(buff));
+			bytes_read = read(fd, cpio->buff, cpio->buff_size);
 		}
 	}
 
@@ -492,9 +625,6 @@ cleanup:
 		fprintf(stderr,"\n");
 	if (fd >= 0)
 		close(fd);
-	if (entry != NULL)
-		archive_entry_free(entry);
-
 	return (0);
 }
 
@@ -695,49 +825,18 @@ mode_list(struct cpio *cpio)
 	exit(0);
 }
 
-/*
- * TODO: Fix hardlink handling.
- */
-static int
-pass_file(struct cpio *cpio, const char *pathname)
-{
-	size_t len;
-	const char *destpath;
-
-	len = strlen(cpio->pass_destdir) + strlen(pathname) + 8;
-	if (len >= cpio->pass_destpath_alloc) {
-		while (len >= cpio->pass_destpath_alloc) {
-			cpio->pass_destpath_alloc += 512;
-			cpio->pass_destpath_alloc *= 2;
-		}
-		free(cpio->pass_destpath);
-		cpio->pass_destpath = malloc(cpio->pass_destpath_alloc);
-		if (cpio->pass_destpath == NULL)
-			cpio_errc(1, ENOMEM,
-			    "Can't allocate path buffer");
-	}
-	if (pathname[0] == '/')
-		pathname++;
-	strcpy(cpio->pass_destpath, cpio->pass_destdir);
-	strcat(cpio->pass_destpath, pathname);
-	destpath = cpio->pass_destpath;
-	if (cpio->option_rename)
-		destpath = cpio_rename(destpath);
-	if (destpath != NULL)
-		file_to_archive(cpio, pathname, destpath);
-	return (0);
-}
-
 static void
 mode_pass(struct cpio *cpio, const char *destdir)
 {
+	struct line_reader *lr;
+	const char *p;
 	int r;
 
 	/* Ensure target dir has a trailing '/' to simplify path surgery. */
-	cpio->pass_destdir = malloc(strlen(destdir) + 8);
-	strcpy(cpio->pass_destdir, destdir);
+	cpio->destdir = malloc(strlen(destdir) + 8);
+	strcpy(cpio->destdir, destdir);
 	if (destdir[strlen(destdir) - 1] != '/')
-		strcat(cpio->pass_destdir, "/");
+		strcat(cpio->destdir, "/");
 
 	cpio->archive = archive_write_disk_new();
 	if (cpio->archive == NULL)
@@ -747,7 +846,10 @@ mode_pass(struct cpio *cpio, const char *destdir)
 		cpio_errc(1, 0, archive_error_string(cpio->archive));
 	cpio->linkresolver = archive_entry_linkresolver_new();
 	archive_write_disk_set_standard_lookup(cpio->archive);
-	process_lines(cpio, "-", pass_file);
+	lr = process_lines_init("-", cpio->line_separator);
+	while ((p = process_lines_next(lr)) != NULL)
+		file_to_archive(cpio, p);
+	process_lines_free(lr);
 
 	archive_entry_linkresolver_free(cpio->linkresolver);
 	r = archive_write_close(cpio->archive);
@@ -806,79 +908,119 @@ cpio_rename(const char *name)
  * terminated with newlines.
  *
  * This uses a self-sizing buffer to handle arbitrarily-long lines.
- * If the "process" function returns non-zero for any line, this
- * function will return non-zero after attempting to process all
- * remaining lines.
  */
-static int
-process_lines(struct cpio *cpio, const char *pathname,
-    int (*process)(struct cpio *, const char *))
-{
+struct line_reader {
 	FILE *f;
 	char *buff, *buff_end, *line_start, *line_end, *p;
-	size_t buff_length, bytes_read, bytes_wanted;
+	char *pathname;
+	size_t buff_length;
 	int separator;
 	int ret;
+};
 
-	separator = cpio->option_null ? '\0' : '\n';
-	ret = 0;
+struct line_reader *
+process_lines_init(const char *pathname, char separator)
+{
+	struct line_reader *lr;
+
+	lr = calloc(1, sizeof(*lr));
+	if (lr == NULL)
+		cpio_errc(1, ENOMEM, "Can't open %s", pathname);
+
+	lr->separator = separator;
+	lr->pathname = strdup(pathname);
 
 	if (strcmp(pathname, "-") == 0)
-		f = stdin;
+		lr->f = stdin;
 	else
-		f = fopen(pathname, "r");
-	if (f == NULL)
+		lr->f = fopen(pathname, "r");
+	if (lr->f == NULL)
 		cpio_errc(1, errno, "Couldn't open %s", pathname);
-	buff_length = 8192;
-	buff = malloc(buff_length);
-	if (buff == NULL)
+	lr->buff_length = 8192;
+	lr->buff = malloc(lr->buff_length);
+	if (lr->buff == NULL)
 		cpio_errc(1, ENOMEM, "Can't read %s", pathname);
-	line_start = line_end = buff_end = buff;
+	lr->line_start = lr->line_end = lr->buff_end = lr->buff;
+
+	return (lr);
+}
+
+const char *
+process_lines_next(struct line_reader *lr)
+{
+	size_t bytes_wanted, bytes_read, new_buff_size;
+	char *line_start, *p;
+
 	for (;;) {
-		/* Get some more data into the buffer. */
-		bytes_wanted = buff + buff_length - buff_end;
-		bytes_read = fread(buff_end, 1, bytes_wanted, f);
-		buff_end += bytes_read;
-		/* Process all complete lines in the buffer. */
-		while (line_end < buff_end) {
-			if (*line_end == separator) {
-				*line_end = '\0';
-				if ((*process)(cpio, line_start) != 0)
-					ret = -1;
-				line_start = line_end + 1;
-				line_end = line_start;
+		/* If there's a line in the buffer, return it immediately. */
+		while (lr->line_end < lr->buff_end) {
+			if (*lr->line_end == lr->separator) {
+				*lr->line_end = '\0';
+				line_start = lr->line_start;
+				lr->line_start = lr->line_end + 1;
+				lr->line_end = lr->line_start;
+				return (line_start);
 			} else
-				line_end++;
+				lr->line_end++;
 		}
-		if (feof(f))
-			break;
-		if (ferror(f))
-			cpio_errc(1, errno, "Can't read %s", pathname);
-		if (line_start > buff) {
+
+		/* If we're at end-of-file, process the final data. */
+		if (lr->f == NULL) {
+			/* If there's more text, return one last line. */
+			if (lr->line_end > lr->line_start) {
+				*lr->line_end = '\0';
+				line_start = lr->line_start;
+				lr->line_start = lr->line_end + 1;
+				lr->line_end = lr->line_start;
+				return (line_start);
+			}
+			/* Otherwise, we're done. */
+			return (NULL);
+		}
+
+		/* Buffer only has part of a line. */
+		if (lr->line_start > lr->buff) {
 			/* Move a leftover fractional line to the beginning. */
-			memmove(buff, line_start, buff_end - line_start);
-			buff_end -= line_start - buff;
-			line_end -= line_start - buff;
-			line_start = buff;
+			memmove(lr->buff, lr->line_start,
+			    lr->buff_end - lr->line_start);
+			lr->buff_end -= lr->line_start - lr->buff;
+			lr->line_end -= lr->line_start - lr->buff;
+			lr->line_start = lr->buff;
 		} else {
 			/* Line is too big; enlarge the buffer. */
-			p = realloc(buff, buff_length *= 2);
+			new_buff_size = lr->buff_length * 2;
+			if (new_buff_size <= lr->buff_length)
+				cpio_errc(1, ENOMEM,
+				    "Line too long in %s", lr->pathname);
+			lr->buff_length = new_buff_size;
+			p = realloc(lr->buff, new_buff_size);
 			if (p == NULL)
 				cpio_errc(1, ENOMEM,
-				    "Line too long in %s", pathname);
-			buff_end = p + (buff_end - buff);
-			line_end = p + (line_end - buff);
-			line_start = buff = p;
+				    "Line too long in %s", lr->pathname);
+			lr->buff_end = p + (lr->buff_end - lr->buff);
+			lr->line_end = p + (lr->line_end - lr->buff);
+			lr->line_start = lr->buff = p;
+		}
+
+		/* Get some more data into the buffer. */
+		bytes_wanted = lr->buff + lr->buff_length - lr->buff_end;
+		bytes_read = fread(lr->buff_end, 1, bytes_wanted, lr->f);
+		lr->buff_end += bytes_read;
+
+		if (ferror(lr->f))
+			cpio_errc(1, errno, "Can't read %s", lr->pathname);
+		if (feof(lr->f)) {
+			if (lr->f != stdin)
+				fclose(lr->f);
+			lr->f = NULL;
 		}
 	}
-	/* At end-of-file, handle the final line. */
-	if (line_end > line_start) {
-		*line_end = '\0';
-		if ((*process)(cpio, line_start) != 0)
-			ret = -1;
-	}
-	free(buff);
-	if (f != stdin)
-		fclose(f);
-	return (ret);
+}
+
+void
+process_lines_free(struct line_reader *lr)
+{
+	free(lr->buff);
+	free(lr->pathname);
+	free(lr);
 }
