@@ -26,7 +26,7 @@
  */
 
 #include "archive_platform.h"
-__FBSDID("$FreeBSD: src/lib/libarchive/archive_write_set_format_ar.c,v 1.5 2008/01/31 08:11:01 kaiw Exp $");
+__FBSDID("$FreeBSD: src/lib/libarchive/archive_write_set_format_ar.c,v 1.7 2008/05/26 17:00:23 kientzle Exp $");
 
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
@@ -87,8 +87,8 @@ archive_write_set_format_ar_bsd(struct archive *_a)
 	struct archive_write *a = (struct archive_write *)_a;
 	int r = archive_write_set_format_ar(a);
 	if (r == ARCHIVE_OK) {
-		a->archive_format = ARCHIVE_FORMAT_AR_BSD;
-		a->archive_format_name = "ar (BSD)";
+		a->archive.archive_format = ARCHIVE_FORMAT_AR_BSD;
+		a->archive.archive_format_name = "ar (BSD)";
 	}
 	return (r);
 }
@@ -99,8 +99,8 @@ archive_write_set_format_ar_svr4(struct archive *_a)
 	struct archive_write *a = (struct archive_write *)_a;
 	int r = archive_write_set_format_ar(a);
 	if (r == ARCHIVE_OK) {
-		a->archive_format = ARCHIVE_FORMAT_AR_GNU;
-		a->archive_format_name = "ar (GNU/SVR4)";
+		a->archive.archive_format = ARCHIVE_FORMAT_AR_GNU;
+		a->archive.archive_format_name = "ar (GNU/SVR4)";
 	}
 	return (r);
 }
@@ -142,12 +142,15 @@ archive_write_ar_header(struct archive_write *a, struct archive_entry *entry)
 	struct ar_w *ar;
 	const char *pathname;
 	const char *filename;
+	int64_t size;
 
 	ret = 0;
 	append_fn = 0;
 	ar = (struct ar_w *)a->format_data;
 	ar->is_strtab = 0;
 	filename = NULL;
+	size = archive_entry_size(entry);
+
 
 	/*
 	 * Reject files with empty name.
@@ -204,7 +207,7 @@ archive_write_ar_header(struct archive_write *a, struct archive_entry *entry)
 		return (ARCHIVE_WARN);
 	}
 
-	if (a->archive_format == ARCHIVE_FORMAT_AR_GNU) {
+	if (a->archive.archive_format == ARCHIVE_FORMAT_AR_GNU) {
 		/*
 		 * SVR4/GNU variant use a "/" to mark then end of the filename,
 		 * make it possible to have embedded spaces in the filename.
@@ -261,7 +264,7 @@ archive_write_ar_header(struct archive_write *a, struct archive_entry *entry)
 				return (ARCHIVE_WARN);
 			}
 		}
-	} else if (a->archive_format == ARCHIVE_FORMAT_AR_BSD) {
+	} else if (a->archive.archive_format == ARCHIVE_FORMAT_AR_BSD) {
 		/*
 		 * BSD variant: for any file name which is more than
 		 * 16 chars or contains one or more embedded space(s), the
@@ -285,8 +288,7 @@ archive_write_ar_header(struct archive_write *a, struct archive_entry *entry)
 				return (ARCHIVE_WARN);
 			}
 			append_fn = 1;
-			archive_entry_set_size(entry,
-			    archive_entry_size(entry) + strlen(filename));
+			size += strlen(filename);
 		}
 	}
 
@@ -322,8 +324,7 @@ stat:
 	}
 
 size:
-	if (format_decimal(archive_entry_size(entry), buff + AR_size_offset,
-	    AR_size_size)) {
+	if (format_decimal(size, buff + AR_size_offset, AR_size_size)) {
 		archive_set_error(&a->archive, ERANGE,
 		    "File size out of range");
 		return (ARCHIVE_WARN);
@@ -333,7 +334,7 @@ size:
 	if (ret != ARCHIVE_OK)
 		return (ret);
 
-	ar->entry_bytes_remaining = archive_entry_size(entry);
+	ar->entry_bytes_remaining = size;
 	ar->entry_padding = ar->entry_bytes_remaining % 2;
 
 	if (append_fn > 0) {
